from ._aggregated_stream import AggregatedStream  # noqa: F401
from ._executor import Executor  # noqa: F401
from ._interfaces import (  # noqa: F401
    MessagePartitioner,
    MessageProtocol,
    MessageSerializer,
    StoredMessage,
    StreamPartitionStatistic,
    SubscriptionState,
    SubscriptionStateProvider,
)
from ._message_store import MessageStore, MessageStoreReader  # noqa: F401
from ._storage import Storage  # noqa: F401
from ._subscription import (  # noqa: F401
    Subscription,
    SubscriptionHandler,
    SubscriptionMessage,
)



import collections as _collections
import signal as _signal
from typing import Callable, Dict, List

from ._interfaces import RunOnNotification
from .tools import PgNotificationListener


class Executor:
    def __init__(self, db_dsn: str):
        self._db_dsn = db_dsn
        self.channel_register: Dict[
            str, List[Callable[[], None]]
        ] = _collections.defaultdict(list)

    def register(self, handler: RunOnNotification):
        self.channel_register[handler.notification_channel].append(handler.run)
        return handler

    def run(self):
        listener = PgNotificationListener(
            dsn=self._db_dsn,
            channels=list(self.channel_register),
            ignore_payload=True,
        )

        def stop():
            print("Stopping...")
            listener.stop()

        _signal.signal(_signal.SIGINT, lambda *_: stop())
        _signal.signal(_signal.SIGTERM, lambda *_: stop())

        listener.start()
        print("Started")
        for notification in listener.messages():
            for handler in self.channel_register[notification.channel]:
                handler()



from typing import Any, TypeVar

from .._interfaces import MessageSerializer

T = TypeVar("T")


class PydanticMessageSerializer(MessageSerializer[T]):
    # the real type would be: (self, message_type: Type[T])
    # but this is not supported by mypy (yet)
    def __init__(self, message_type: Any) -> None:
        self.message_type: type[T] = message_type

    def serialize(self, message: T) -> dict:
        # TODO pydantic v1 compatibility
        return message.model_dump(mode="json")  # type: ignore

    def deserialize(self, message: dict) -> T:
        import pydantic as _pydantic

        # TODO pydantic v1 compatibility
        return _pydantic.TypeAdapter(self.message_type).validate_python(message)



import dataclasses as _dc
import typing as _typing
from typing import Callable, Dict, Generic, Iterator, Type, TypeVar, Union

from ._aggregated_stream import AggregatedStream
from ._compat import UNION_TYPES
from ._interfaces import (
    LockProvider,
    MessageProtocol,
    StoredMessage,
    SubscriptionStateProvider,
)

E = TypeVar("E", bound=MessageProtocol)


@_dc.dataclass(frozen=True)
class SubscriptionMessage(Generic[E]):
    partition: int
    position: int
    stored_message: StoredMessage[E]


class Subscription(Generic[E]):
    def __init__(
        self,
        name: str,
        stream: AggregatedStream[E],
        state_provider: SubscriptionStateProvider,
        lock_provider: LockProvider,
        # TODO start at time
        # TODO start at "next message"
    ):
        assert name.isidentifier(), "Group name must be a valid identifier"
        self.name = name
        self._stream = stream
        self._lock_provider = lock_provider
        self._state_provider = state_provider
        self.handler = SubscriptionHandler(self)

    def get_next_messages(self, count: int) -> Iterator[SubscriptionMessage[E]]:
        state = self._state_provider.read(self.name)
        statistics = list(
            self._stream.get_partition_statistics(
                position_limits=state.positions, result_limit=10
            )
        )
        for statistic in statistics:
            lock_key = f"subscription-{self.name}-{statistic.partition_number}"
            if not self._lock_provider.lock(lock_key):
                continue
            # now we have the lock, we need to check if the position is still valid
            # if not, we need to release the lock and try the next partition
            state = self._state_provider.read(self.name)
            if state.positions.get(statistic.partition_number, -1) != (
                statistic.next_message_position - 1
            ):
                self._lock_provider.unlock(lock_key)
                continue
            try:
                with self._stream._store.reader() as reader:
                    message_pointers = list(
                        self._stream.read_slice(
                            partition=statistic.partition_number,
                            start=statistic.next_message_position,
                            count=count,
                        )
                    )
                    stored_messages = {
                        message.message_id: message
                        for message in reader.get_messages_by_ids(
                            [pointer.message_id for pointer in message_pointers]
                        )
                    }

                    for pointer in message_pointers:
                        yield SubscriptionMessage(
                            partition=pointer.partition,
                            position=pointer.position,
                            stored_message=stored_messages[pointer.message_id],
                        )
                        self._ack(
                            partition=pointer.partition,
                            position=pointer.position,
                        )
                break
            finally:
                self._lock_provider.unlock(lock_key)

    def _ack(self, partition: int, position: int):
        state = self._state_provider.read(self.name)
        assert (
            state.positions.get(partition, -1) == position - 1
        ), f"{partition} should have {position - 1} as last position, but has {state.positions.get(partition, -1)}"
        self._state_provider.store(
            subscription_name=self.name, partition=partition, position=position
        )


HandlerCallable = Union[
    Callable[[E], None],
    Callable[[StoredMessage[E]], None],
    Callable[[SubscriptionMessage[E]], None],
]


@_dc.dataclass
class _Handler:
    handler: HandlerCallable
    pass_subscription_message: bool
    pass_stored_message: bool

    def exec(self, message: SubscriptionMessage):
        if self.pass_subscription_message:
            self.handler(message)  # type: ignore
        elif self.pass_stored_message:
            self.handler(message.stored_message)  # type: ignore
        else:
            self.handler(message.stored_message.message)


H = TypeVar("H", bound=HandlerCallable)


class SubscriptionHandler(Generic[E]):
    def __init__(self, subscription: Subscription[E]):
        self._subscription = subscription
        self._handlers: Dict[Type[E], _Handler] = {}
        self._batch_size = 1

    @property
    def notification_channel(self) -> str:
        return self._subscription._stream.notification_channel

    def run(self):
        assert (
            self._subscription.handler is self
        ), "A subscription can only have one handler"
        self.run_once()

    def register(self, handler: H) -> H:
        if len(handler.__annotations__) != 1:
            raise ValueError(
                "Handler must accept exactly one (type annotated) argument"
            )

        pass_subscription_message = False
        pass_stored_message = False
        handled_type = list(handler.__annotations__.values())[0]
        if str(handled_type).startswith(
            "depeche_db._subscription.SubscriptionMessage["
        ):
            pass_subscription_message = True
            handled_type = handled_type.__args__[0]
        if str(handled_type).startswith("depeche_db._interfaces.StoredMessage["):
            pass_stored_message = True
            handled_type = handled_type.__args__[0]

        self.assert_not_registered(handled_type)
        self._handlers[handled_type] = _Handler(
            handler=handler,
            pass_subscription_message=pass_subscription_message,
            pass_stored_message=pass_stored_message,
        )
        return handler

    def assert_not_registered(self, handled_type: Type[E]):
        if _typing.get_origin(handled_type) in UNION_TYPES:
            for member in _typing.get_args(handled_type):
                self.assert_not_registered(member)
        else:
            for registered_type in self._handlers:
                if issubclass(handled_type, registered_type):
                    raise ValueError(
                        f"Handler for {handled_type} is already registered for {registered_type}"
                    )

    def handle(self, message: SubscriptionMessage):
        message_type = type(message.stored_message.message)
        for handled_type, handler in self._handlers.items():
            if issubclass(message_type, handled_type):
                try:
                    handler.exec(message)
                except Exception:
                    # TODO error handler!
                    pass
                return

    def run_once(self):
        while True:
            n = 0
            for message in self._subscription.get_next_messages(count=self._batch_size):
                n += 1
                self.handle(message)
            if n == 0:
                break



import sys
import typing

if sys.version_info < (3, 9):
    raise RuntimeError("DepecheDB requires Python 3.9 or newer")
elif sys.version_info < (3, 10):
    UNION_TYPES = (typing.Union,)
else:
    import types

    UNION_TYPES = (typing.Union, types.UnionType)


SA_VERSION = "2.x"
try:
    from sqlalchemy import Connection as SAConnection  # noqa
except ImportError:
    SA_VERSION = "1.4.x"
    from sqlalchemy.engine import Connection as SAConnection  # noqa



import uuid as _uuid
from typing import Iterator, Sequence, Tuple

import sqlalchemy as _sa
from sqlalchemy_utils import UUIDType as _UUIDType

from ._compat import SAConnection
from ._interfaces import MessagePosition


class Storage:
    name: str

    def __init__(self, name: str, engine: _sa.engine.Engine):
        assert name.isidentifier(), "name must be a valid identifier"
        self.name = name
        self.metadata = _sa.MetaData()
        self.message_table = _sa.Table(
            f"{name}_messages",
            self.metadata,
            _sa.Column("message_id", _UUIDType(), primary_key=True),
            _sa.Column(
                "global_position",
                _sa.Integer,
                _sa.Sequence(f"{name}_messages_global_position_seq"),
                unique=True,
                nullable=False,
            ),
            _sa.Column(
                "added_at", _sa.DateTime, nullable=False, server_default=_sa.func.now()
            ),
            _sa.Column("stream", _sa.String(255), nullable=False),
            _sa.Column("version", _sa.Integer, nullable=False),
            _sa.Column("message", _sa.JSON, nullable=False),
            _sa.UniqueConstraint(
                "stream", "version", name=f"{name}_stream_version_unique"
            ),
        )
        self.notification_channel = f"{name}_messages"
        trigger = _sa.DDL(
            f"""
            CREATE OR REPLACE FUNCTION {name}_notify_message_inserted()
              RETURNS trigger AS $$
            DECLARE
            BEGIN
              PERFORM pg_notify(
                '{self.notification_channel}',
                json_build_object(
                    'message_id', NEW.message_id,
                    'stream', NEW.stream,
                    'version', NEW.version,
                    'global_position', NEW.global_position
                )::text);
              RETURN NEW;
            END;
            $$ LANGUAGE plpgsql;

            CREATE TRIGGER {name}_notify_message_inserted
              AFTER INSERT ON {name}_messages
              FOR EACH ROW
              EXECUTE PROCEDURE {name}_notify_message_inserted();
            """
        )
        _sa.event.listen(
            self.message_table, "after_create", trigger.execute_if(dialect="postgresql")
        )
        self.metadata.create_all(engine, checkfirst=True)

    def add(
        self,
        conn: SAConnection,
        stream: str,
        expected_version: int,
        message_id: _uuid.UUID,
        message: dict,
    ) -> MessagePosition:
        return self.add_all(conn, stream, expected_version, [(message_id, message)])

    def add_all(
        self,
        conn: SAConnection,
        stream: str,
        expected_version: int,
        messages: Sequence[Tuple[_uuid.UUID, dict]],
    ) -> MessagePosition:
        max_version = self.get_max_version(conn, stream).version
        if expected_version > -1:
            if max_version != expected_version:
                raise ValueError("optimistic concurrency failure")
        conn.execute(
            self.message_table.insert(),
            [
                {
                    "message_id": message_id,
                    "stream": stream,
                    "version": max_version + i + 1,
                    "message": message,
                }
                for i, (message_id, message) in enumerate(messages)
            ],
        )
        return self.get_max_version(conn, stream)

    def get_max_version(self, conn: SAConnection, stream: str) -> MessagePosition:
        row = conn.execute(
            _sa.select(
                _sa.func.max(self.message_table.c.version).label("version"),
                _sa.func.max(self.message_table.c.global_position).label(
                    "global_position"
                ),
            )
            .select_from(self.message_table)
            .where(self.message_table.c.stream == stream),
        ).fetchone()
        if not row or row.version is None:
            return MessagePosition(stream, 0, None)
        return MessagePosition(stream, row.version, row.global_position)

    def get_message_ids(self, conn: SAConnection, stream: str) -> Iterator[_uuid.UUID]:
        for id in conn.execute(
            _sa.select(self.message_table.c.message_id)
            .select_from(self.message_table)
            .where(self.message_table.c.stream == stream)
            .order_by(self.message_table.c.version)
        ).scalars():
            yield id

    def read(
        self, conn: SAConnection, stream: str
    ) -> Iterator[Tuple[_uuid.UUID, int, dict, int]]:
        return conn.execute(  # type: ignore
            _sa.select(
                self.message_table.c.message_id,
                self.message_table.c.version,
                self.message_table.c.message,
                self.message_table.c.global_position,
            )
            .select_from(self.message_table)
            .where(self.message_table.c.stream == stream)
            .order_by(self.message_table.c.version)
        )

    def read_multiple(
        self, conn: SAConnection, streams: Sequence[str]
    ) -> Iterator[Tuple[_uuid.UUID, str, int, dict, int]]:
        return conn.execute(  # type: ignore
            _sa.select(
                self.message_table.c.message_id,
                self.message_table.c.stream,
                self.message_table.c.version,
                self.message_table.c.message,
                self.message_table.c.global_position,
            )
            .select_from(self.message_table)
            .where(self.message_table.c.stream.in_(streams))
            .order_by(self.message_table.c.global_position)
        )

    def read_wildcard(
        self, conn: SAConnection, stream_wildcard: str
    ) -> Iterator[Tuple[_uuid.UUID, str, int, dict, int]]:
        return conn.execute(  # type: ignore
            _sa.select(
                self.message_table.c.message_id,
                self.message_table.c.stream,
                self.message_table.c.version,
                self.message_table.c.message,
                self.message_table.c.global_position,
            )
            .select_from(self.message_table)
            .where(self.message_table.c.stream.like(stream_wildcard))
            .order_by(self.message_table.c.global_position)
        )

    def get_message_by_id(
        self, conn: SAConnection, message_id: _uuid.UUID
    ) -> Tuple[_uuid.UUID, str, int, dict, int]:
        return conn.execute(  # type: ignore
            _sa.select(
                self.message_table.c.message_id,
                self.message_table.c.stream,
                self.message_table.c.version,
                self.message_table.c.message,
                self.message_table.c.global_position,
            ).where(self.message_table.c.message_id == message_id)
        ).first()

    def get_messages_by_ids(
        self, conn: SAConnection, message_ids: Sequence[_uuid.UUID]
    ) -> Iterator[Tuple[_uuid.UUID, str, int, dict, int]]:
        return conn.execute(  # type: ignore
            _sa.select(
                self.message_table.c.message_id,
                self.message_table.c.stream,
                self.message_table.c.version,
                self.message_table.c.message,
                self.message_table.c.global_position,
            ).where(self.message_table.c.message_id.in_(message_ids))
        )

    def truncate(self, conn: SAConnection):
        conn.execute(self.message_table.delete())
